/***************************************************************************/
/* free.c								   */
/*									   */
/* Free storage for the structures defined in spki.h			   */
/***************************************************************************/

/***************************************************************************/
/*									   */
/*  Rules for refcnt handling.						   */
/*									   */
/*  Any free block that can be linked off another free block contains	   */
/*  a refcnt.  Such a block, when carved, has a refcnt == 1, to indicate   */
/*  the usage on the stack (in an automatic variable or in the return	   */
/*  from the new_xxx function).  Any routine that calls new_xxx should,	   */
/*  at the end of the routine, call drop_xxx to show that the stack	   */
/*  reference has gone away.						   */
/*									   */
/*  When storing such a pointer in any structure, one should use	   */
/*  store_xxx calls, which correctly adjust the refcnt values of	   */
/*  the block being put away and any block that used to be in that	   */
/*  location.								   */
/*									   */
/*  Note: the hash table blocks (Ren5 and Tred) do not have refcnt	   */
/*  since they are not referred to by other structures, except for	   */
/*  simple chaining.							   */
/*									   */
/***************************************************************************/


#include <stdlib.h>
#include <string.h>
#include "abbrev.h"
#include "sexp.h"
#include "free.h"
#include "tuple.h"
#include "sutil.h"

static pBytes bytesfree = NULL ; /* byte strings */
static pSexp sexpfree = NULL ;	/* S-expressions */
static pTuple tuplefree = NULL ; /* tuples */
static pRen5 ren5free = NULL ;	/* Ren5 blocks */
static pPkh pkhfree = NULL ;	/* Pkh blocks */
static pTred tredfree = NULL ;	/* Threshold reduction blocks */
static pThresh threshfree = NULL ; /* Threshold rendezvous blocks */
static pRenlists renlistsfree = NULL ; /* rendezvous list blocks */

/***************************************************************************/
/* new_bytes								   */
/*									   */
/* Carve and return a new pBytes block					   */
/***************************************************************************/

pBytes new_bytes( void )
{
  pBytes h = bytesfree ;		/* try for one from the list */

  if (h == NULL) {
    h = (pBytes) malloc( sizeof ( Bytes ) ) ;
    if (h == NULL) return ( NULL ) ;
    memset( h, 0, sizeof( Bytes ) ) ;
  } else {
    bytesfree = h->nxt ;
    h->nxt = NULL ;
  }
  h->refcnt = 1 ;		/* one for the stack reference */

  return ( h ) ;
} /* new_bytes */

/***************************************************************************/
/* drop_bytes								   */
/*									   */
/* Drop h, returning any internal storage and put it on the free list.	   */
/***************************************************************************/

void drop_bytes( pBytes h )
{
  if (h == NULL) return ;
  if ((--(h->refcnt)) > 0) return ;
  if (h->drop_cont) {
    if (h->disp != NULL) free( h->disp ) ;
    if (h->val != NULL) free( h->val ) ;
  } /* inner byte strings were carved by malloc */
  memset( h, 0, sizeof( Bytes ) ) ;
  /* chain it on bytesfree */
  h->nxt = bytesfree ;
  bytesfree = h ;
  return ;
} /* drop_bytes */

/***************************************************************************/
/* new_sexp								   */
/*									   */
/* Carve and return a new pSexp block					   */
/***************************************************************************/

pSexp new_sexp( void )
{
  pSexp h = sexpfree ;		/* try for one from the list */

  if (h == NULL) {
    h = (pSexp) malloc( sizeof ( Sexp ) ) ;
    if (h == NULL) return ( NULL ) ;
    memset( h, 0, sizeof( Sexp ) ) ;
  } else {
    sexpfree = h->nxt ;
    h->nxt = NULL ;
  }
  h->refcnt = 1 ;		/* one for the stack reference */

  return ( h ) ;
} /* new_sexp */

/***************************************************************************/
/* drop_sexp								   */
/*									   */
/* Drop h, returning any internal storage and put it on the free list.	   */
/***************************************************************************/

void drop_sexp( pSexp h )
{
  if (h == NULL) return ;
  if ((--(h->refcnt)) > 0) return ;
  /* what to drop depends on what kind of node this is */
  switch (h->typ) {
  case S_UKW: case S_BS:	/* byte string to free */
    drop_bytes( h->u.b ) ;	/* drop it */
    break ;

  case S_SEXP:			/* S-expression to free */
    drop_sexp( h->u.s ) ;	/* drop it */
    break ;

  case S_KW:
  default: /* no action */
  } /* switch */
  drop_sexp( h->term ) ;	/* drop the remainder of the expression */

  memset( h, 0, sizeof( Sexp ) ) ; /* all dropped, clear for good measure */
  /* chain it on sexpfree */
  h->nxt = sexpfree ;
  sexpfree = h ;
  return ;
} /* drop_sexp */

/***************************************************************************/
/* new_tred								   */
/*									   */
/* Carve and return a new pTred block					   */
/***************************************************************************/

pTred new_tred( void )
{
  pTred h = tredfree ;		/* try for one from the list */

  if (h == NULL) {
    h = (pTred) malloc( sizeof ( Tred ) ) ;
    if (h == NULL) return ( NULL ) ;
    memset( h, 0, sizeof( Tred ) ) ;
  } else {
    tredfree = h->nxt ;
    h->nxt = NULL ;
  }

  return ( h ) ;
} /* new_tred */

/***************************************************************************/
/* drop_tred								   */
/*									   */
/* Drop h, returning any internal storage and put it on the free list.	   */
/***************************************************************************/

void drop_tred( pTred h )
{
  if (h == NULL) return ;
  drop_renlists( h->s ) ;	/* drop the Renlists block */
  /* don't drop h->th -- since we didn't count the refcnt for this entry */
  /* don't drop h->t either, since we didn't refcnt it */
  memset( h, 0, sizeof( Tred ) ) ; /* clear the block */
				/* and chain it on tredfree */
  h->nxt = tredfree ;
  tredfree = h ;
  return ;			/* and return */
} /* drop_tred */

/***************************************************************************/
/* drop_tred_list							   */
/*									   */
/* Drop an entire list, linked by .nxt					   */
/***************************************************************************/

void drop_tred_list( pTred c )
{
  pTred n = c ;			/* for the next pointer */

  while ((c = n) != NULL) {
    n = c->nxt ;		/* next pointer */
    drop_tred( c ) ;
  } /* while */
} /* drop_tred_list */

/***************************************************************************/
/* new_ren5								   */
/*									   */
/* Carve and return a new pRen5 block					   */
/***************************************************************************/

pRen5 new_ren5( void )
{
  pRen5 h = ren5free ;		/* try for one from the list */

  if (h == NULL) {
    h = (pRen5) malloc( sizeof ( Ren5 ) ) ;
    if (h == NULL) return ( NULL ) ;
    memset( h, 0, sizeof( Ren5 ) ) ;
  } else {
    ren5free = h->nxt ;
    h->nxt = NULL ;
  }

  return ( h ) ;
} /* new_ren5 */

/***************************************************************************/
/* drop_ren5								   */
/*									   */
/* Drop h, returning any internal storage and put it on the free list.	   */
/***************************************************************************/

void drop_ren5( pRen5 h )
{
  if (h == NULL) return ;
  drop_tuple( h->t ) ;		/* and any Tuple reference */
  memset( h, 0, sizeof( Ren5 ) ) ; /* clear the block, to be safe */
				/* and chain it on ren5free */
  h->nxt = ren5free ;
  ren5free = h ;
  return ;			/* and return */
} /* drop_ren5 */

/***************************************************************************/
/* drop_ren5_list							   */
/*									   */
/* Drop an entire list, linked by .nxt					   */
/***************************************************************************/

void drop_ren5_list( pRen5 c )
{
  pRen5 n = c ;			/* for the next pointer */

  while ((c = n) != NULL) {
    n = c->nxt ;		/* next pointer */
    drop_ren5( c ) ;
  } /* while */
} /* drop_ren5_list */

/***************************************************************************/
/* new_pkh								   */
/*									   */
/* Carve and return a new pPkh block					   */
/***************************************************************************/

pPkh new_pkh( void )
{
  pPkh h = pkhfree ;		/* try for one from the list */

  if (h == NULL) {
    h = (pPkh) malloc( sizeof ( Pkh ) ) ;
    if (h == NULL) return ( NULL ) ;
    memset( h, 0, sizeof( Pkh ) ) ;
  } else {
    pkhfree = h->nxt ;
    h->nxt = NULL ;
  }

  return ( h ) ;
} /* new_pkh */

/***************************************************************************/
/* drop_pkh								   */
/*									   */
/* Drop h, returning any internal storage and put it on the free list.	   */
/***************************************************************************/

void drop_pkh( pPkh h )
{
  if (h == NULL) return ;
  drop_renlists( h->obj ) ;	/* drop the object reference */
  memset( h, 0, sizeof( Pkh ) ) ; /* clear the block, to be safe */
				/* and chain it on pkhfree */
  h->nxt = pkhfree ;
  pkhfree = h ;
  return ;			/* and return */
} /* drop_pkh */

/***************************************************************************/
/* drop_pkh_list							   */
/*									   */
/* Drop an entire list, linked by .nxt					   */
/***************************************************************************/

void drop_pkh_list( pPkh c )
{
  pPkh n = c ;			/* for the next pointer */

  while ((c = n) != NULL) {
    n = c->nxt ;		/* next pointer */
    drop_pkh( c ) ;
  } /* while */
} /* drop_pkh_list */

/***************************************************************************/
/* new_tuple								   */
/*									   */
/* Carve and return a new pTuple block					   */
/***************************************************************************/

pTuple new_tuple( void )
{
  pTuple h = tuplefree ;		/* try for one from the list */

  if (h == NULL) {
    h = (pTuple) malloc( sizeof ( Tuple ) ) ;
    if (h == NULL) return ( NULL ) ;
    memset( h, 0, sizeof( Tuple ) ) ;
  } else {
    tuplefree = h->nxt ;
    h->nxt = NULL ;
  }
  h->refcnt = 1 ;		/* one for the stack reference */

  return ( h ) ;
} /* new_tuple */

/***************************************************************************/
/* drop_tuple								   */
/*									   */
/* Drop h, returning any internal storage and put it on the free list.	   */
/***************************************************************************/

void drop_tuple( pTuple h )
{
  if (h == NULL) return ;
  if ((--(h->refcnt)) > 0) return ;
  /* yes, it's time to drop this for real */
  drop_renlists( h->ik ) ;	/* issuer key */
  drop_bytes( h->inam ) ;	/* issuer name */
  drop_renlists( h->sk ) ;	/* subject key */
  drop_bytes( h->snam ) ;	/* subject name */
  drop_sexp( h->rnames ) ;	/* rest of the subject names */
  drop_thresh( h->th ) ;	/* any threshold "issuer" */
  drop_sexp( h->s ) ;		/* drop the subject */
  drop_sexp( h->t ) ;		/* and the tag */
  drop_bytes( h->nb ) ;		/* and the not-before date.time */
  drop_bytes( h->na ) ;		/* and the not-after */

  /* done dropping references to structures */
  memset( h, 0, sizeof( Tuple ) ) ; /* clear the memory, to be safe */
				/* and chain it on tuplefree */
  h->nxt = tuplefree ;
  tuplefree = h ;
  return ;			/* and return */
} /* drop_tuple */

/***************************************************************************/
/* new_thresh								   */
/*									   */
/* Carve and return a new pThresh block					   */
/***************************************************************************/

pThresh new_thresh( void )
{
  pThresh h = threshfree ;		/* try for one from the list */

  if (h == NULL) {
    h = (pThresh) malloc( sizeof ( Thresh ) ) ;
    if (h == NULL) return ( NULL ) ;
    memset( h, 0, sizeof( Thresh ) ) ;
  } else {
    threshfree = h->nxt ;
    h->nxt = NULL ;
  }
  h->refcnt = 1 ;		/* one for the stack reference */

  return ( h ) ;
} /* new_thresh */

/***************************************************************************/
/* drop_thresh								   */
/*									   */
/* Drop h, returning any internal storage and put it on the free list.	   */
/***************************************************************************/

void drop_thresh( pThresh h )
{
  int i ;			/* loop index */

  if (h == NULL) return ;
  if ((--(h->refcnt)) > 0) return ;

  drop_renlists( h->ik ) ;	/* issuer key Renlists */
  drop_thresh( h->th ) ;	/* possible "issuer" threshold */
  if (h->tr != NULL) {
    for (i=0; i<=h->n; i++)
      drop_tred_list( h->tr[i] ) ; /* list of tuples waiting to reduce */
    free( h->tr ) ;		/* free the array itself */
  } /* if */
  memset( h, 0, sizeof( Thresh ) ) ; /* clear all memory, to be safe */
				/* and chain it on threshfree */
  h->nxt = threshfree ;
  threshfree = h ;
  return ;			/* and return */
} /* drop_thresh */

/***************************************************************************/
/* store_sexp								   */
/*									   */
/* Store an sexp pointer in a structure, adjusting the refcnts.		   */
/***************************************************************************/

void store_sexp( pSexp *dest, pSexp src )
{
  if (dest == NULL) {
    drop_sexp( src ) ;		/* drop what we're putting away */
    return ;
  } /* no destination */
  if (src != NULL)		/* bump first, in case src == *dest */
    src->refcnt++ ;		/* bump its reference count, for the */
				/* coming storage into *dest */
  drop_sexp( *dest ) ;		/* drop whatever is there now */
  *dest = src ;			/* store the new value */
  return ;
} /* store_sexp */

/***************************************************************************/
/* store_bytes								   */
/*									   */
/* Store a bytes pointer in a structure, adjusting the refcnts. 	   */
/***************************************************************************/

void store_bytes( pBytes *dest, pBytes src )
{
  if (dest == NULL) {
    drop_bytes( src ) ;		/* drop what we're putting away */
    return ;
  } /* no destination */
  if (src != NULL)		/* bump first, in case src == *dest */
    src->refcnt++ ;		/* bump its reference count */
  drop_bytes( *dest ) ;		/* drop whatever is there */
  *dest = src ;			/* store the new value */
  return ;
} /* store_bytes */

/***************************************************************************/
/* store_tuple								   */
/*									   */
/* Store a tuple pointer in a structure, adjusting the refcnts. 	   */
/***************************************************************************/

void store_tuple( pTuple *dest, pTuple src )
{
  if (dest == NULL) {
    drop_tuple( src ) ;		/* drop what we're putting away */
    return ;
  } /* no destination */
  if (src != NULL)		/* bump first, in case src == *dest */
    src->refcnt++ ;		/* bump its reference count */
  drop_tuple( *dest ) ;		/* drop whatever is there */
  *dest = src ;			/* store the new value */
  return ;
} /* store_tuple */

/***************************************************************************/
/* store_thresh								   */
/*									   */
/* Store a thresh pointer in a structure, adjusting the refcnts. 	   */
/***************************************************************************/

void store_thresh( pThresh *dest, pThresh src )
{
  if (dest == NULL) {
    drop_thresh( src ) ;	/* drop what we're putting away */
    return ;
  } /* no destination */
  if (src != NULL)		/* bump first, in case src == *dest */
    src->refcnt++ ;		/* bump its reference count */
  drop_thresh( *dest ) ;	/* drop whatever is there */
  *dest = src ;			/* store the new value */
  return ;
} /* store_thresh */

/***************************************************************************/
/* new_renlists								   */
/*									   */
/* Carve and return a new pRenlists block 				   */
/***************************************************************************/

pRenlists new_renlists( void )
{
  pRenlists h = renlistsfree ;	/* try for one from the list */

  if (h == NULL) {
    h = (pRenlists) malloc( sizeof ( Renlists ) ) ;
    if (h == NULL) return ( NULL ) ;
    memset( h, 0, sizeof( Renlists ) ) ;
  } else {
    renlistsfree = h->nxt ;
    h->nxt = NULL ;
  }
  h->refcnt = 1 ;		/* one for the stack reference */

  return ( h ) ;
} /* new_renlists */

/***************************************************************************/
/* drop_renlists 							   */
/*									   */
/* Drop h, returning any internal storage and put it on the free list. 	   */
/***************************************************************************/

void drop_renlists( pRenlists h )
{
  if (h == NULL) return ;
  if ((--(h->refcnt)) > 0) return ;

  drop_ren5_list( h->i ) ;
  drop_ren5_list( h->s ) ;

  memset( h, 0, sizeof( Renlists ) ) ; /* clear all memory, to be safe */
				/* and chain it on renlistsfree */
  h->nxt = renlistsfree ;
  renlistsfree = h ;
  return ;			/* and return */
} /* drop_renlists */

/***************************************************************************/
/* store_renlists 							   */
/*									   */
/* Store a renlists pointer in a structure, adjusting the refcnts. 	   */
/***************************************************************************/

void store_renlists( pRenlists *dest, pRenlists src )
{
  if (dest == NULL) {
    drop_renlists( src ) ;	/* drop what we're putting away */
    return ;
  } /* no destination */
  if (src != NULL)		/* bump first, in case src == *dest */
    src->refcnt++ ;		/* bump its reference count */
  drop_renlists( *dest ) ;	/* drop whatever is there */
  *dest = src ;			/* store the new value */
  return ;
} /* store_renlists */

/***************************************************************************/
/* new_sexp_sexp 							   */
/*									   */
/* Carve and return a new S-expression sub-expression element and fill it  */
/* in with the parameter.						   */
/***************************************************************************/

pSexp new_sexp_sexp( pSexp s )
{
  pSexp elt = new_sexp() ;	/* carve it */

  if (elt == NULL) return ( NULL ) ;
  elt->typ = S_SEXP ;
  store_sexp( &(elt->u.s), s ) ;
  return ( elt ) ;
} /* new_sexp_sexp */

/***************************************************************************/
/* new_sexp_bs								   */
/*									   */
/* Carve and return a new S-expression sub-expression element and fill it  */
/* with the given byte string.						   */
/***************************************************************************/

pSexp new_sexp_bs( pBytes b )
{
  pSexp elt = new_sexp() ;	/* carve it */

  if (elt == NULL) return ( NULL ) ;
  elt->typ = S_BS ;
  store_bytes( &(elt->u.b), b ) ;
  return ( elt ) ;
} /* new_sexp_bs */

/***************************************************************************/
/* new_sexp_kw								   */
/*									   */
/* Carve and return a new keyword S-expression node.			   */
/***************************************************************************/

pSexp new_sexp_kw( int k )
{
  pSexp elt = new_sexp() ;	/* carve it */

  if (elt == NULL) return ( NULL ) ;
  elt->typ = S_KW ;
  elt->u.k = k ;
  return ( elt ) ;
} /* new_sexp_kw */

/***************************************************************************/
/* copy_sexp_elt							   */
/*									   */
/* Copy one S-expression element.					   */
/***************************************************************************/

pSexp copy_sexp_elt( pSexp e )
{
  pSexp res ;			/* the result */
  switch (pull_typ( e ) ) {
  case S_KW: return ( new_sexp_kw( e->u.k ) ) ;
  case S_SEXP: return ( new_sexp_sexp( e->u.s ) ) ;
  case S_UKW: case S_BS:
    res = new_sexp_bs( e->u.b ) ;
    res->typ = e->typ ;
    return ( res ) ;
  default: return ( NULL ) ;
  } /* switch pull_typ( e ) */
} /* copy_sexp_elt */
