/***************************************************************************/
/* free.h								   */
/*									   */
/* Free storage for the structures defined in spki.h			   */
/***************************************************************************/

/***************************************************************************/
/*									   */
/*  Rules for refcnt handling.						   */
/*									   */
/*  Any free block that can be linked off another free block contains	   */
/*  a refcnt.  Such a block, when carved, has a refcnt == 1, to indicate   */
/*  the usage on the stack (in an automatic variable or in the return	   */
/*  from the new_xxx function).  Any routine that calls new_xxx should,	   */
/*  at the end of the routine, call drop_xxx to show that the stack	   */
/*  reference has gone away.						   */
/*									   */
/*  When storing such a pointer in any structure, one should use	   */
/*  store_xxx calls, which correctly adjust the refcnt values of	   */
/*  the block being put away and any block that used to be in that	   */
/*  location.								   */
/*									   */
/*  Note: the hash table blocks (Ren5 and Tred) do not have refcnt	   */
/*  since they are not referred to by other structures, except for	   */
/*  simple chaining.							   */
/*									   */
/***************************************************************************/

#if !defined (__FREE_H__)
#define __FREE_H__

#include "sexp.h"
#include "tuple.h"

/***************************************************************************/
/* new_bytes								   */
/*									   */
/* Carve and return a new pBytes block					   */
/***************************************************************************/

pBytes new_bytes( void ) ;

/***************************************************************************/
/* drop_bytes								   */
/*									   */
/* Drop h, returning any internal storage and put it on the free list.	   */
/***************************************************************************/

void drop_bytes( pBytes h ) ;

/***************************************************************************/
/* new_sexp								   */
/*									   */
/* Carve and return a new pSexp block					   */
/***************************************************************************/

pSexp new_sexp( void ) ;

/***************************************************************************/
/* drop_sexp								   */
/*									   */
/* Drop h, returning any internal storage and put it on the free list.	   */
/***************************************************************************/

void drop_sexp( pSexp h ) ;

/***************************************************************************/
/* new_tred								   */
/*									   */
/* Carve and return a new pTred block					   */
/***************************************************************************/

pTred new_tred( void ) ;

/***************************************************************************/
/* drop_tred								   */
/*									   */
/* Drop h, returning any internal storage and put it on the free list.	   */
/***************************************************************************/

void drop_tred( pTred h ) ;

/***************************************************************************/
/* drop_tred_list							   */
/*									   */
/* Drop an entire list, linked by .nxt					   */
/***************************************************************************/

void drop_tred_list( pTred c ) ;

/***************************************************************************/
/* new_ren5								   */
/*									   */
/* Carve and return a new pRen5 block					   */
/***************************************************************************/

pRen5 new_ren5( void ) ;

/***************************************************************************/
/* drop_ren5								   */
/*									   */
/* Drop h, returning any internal storage and put it on the free list.	   */
/***************************************************************************/

void drop_ren5( pRen5 h ) ;

/***************************************************************************/
/* drop_ren5_list							   */
/*									   */
/* Drop an entire list, linked by .nxt					   */
/***************************************************************************/

void drop_ren5_list( pRen5 c ) ;

/***************************************************************************/
/* new_pkh								   */
/*									   */
/* Carve and return a new pPkh block					   */
/***************************************************************************/

pPkh new_pkh( void ) ;

/***************************************************************************/
/* drop_pkh								   */
/*									   */
/* Drop h, returning any internal storage and put it on the free list.	   */
/***************************************************************************/

void drop_pkh( pPkh h ) ;

/***************************************************************************/
/* drop_pkh_list							   */
/*									   */
/* Drop an entire list, linked by .nxt					   */
/***************************************************************************/

void drop_pkh_list( pPkh c ) ;

/***************************************************************************/
/* new_tuple								   */
/*									   */
/* Carve and return a new pTuple block					   */
/***************************************************************************/

pTuple new_tuple( void ) ;

/***************************************************************************/
/* drop_tuple								   */
/*									   */
/* Drop h, returning any internal storage and put it on the free list.	   */
/***************************************************************************/

void drop_tuple( pTuple h ) ;

/***************************************************************************/
/* new_thresh								   */
/*									   */
/* Carve and return a new pThresh block					   */
/***************************************************************************/

pThresh new_thresh( void ) ;

/***************************************************************************/
/* drop_thresh								   */
/*									   */
/* Drop h, returning any internal storage and put it on the free list.	   */
/***************************************************************************/

void drop_thresh( pThresh h ) ;

/***************************************************************************/
/* store_sexp								   */
/*									   */
/* Store an sexp pointer in a structure, adjusting the refcnts.		   */
/***************************************************************************/

void store_sexp( pSexp *dest, pSexp src ) ;

/***************************************************************************/
/* store_bytes								   */
/*									   */
/* Store a bytes pointer in a structure, adjusting the refcnts. 	   */
/***************************************************************************/

void store_bytes( pBytes *dest, pBytes src ) ;

/***************************************************************************/
/* store_tuple								   */
/*									   */
/* Store a tuple pointer in a structure, adjusting the refcnts. 	   */
/***************************************************************************/

void store_tuple( pTuple *dest, pTuple src ) ;

/***************************************************************************/
/* store_thresh								   */
/*									   */
/* Store a thresh pointer in a structure, adjusting the refcnts. 	   */
/***************************************************************************/

void store_thresh( pThresh *dest, pThresh src ) ;

/***************************************************************************/
/* new_renlists								   */
/*									   */
/* Carve and return a new pRenlists block 				   */
/***************************************************************************/

pRenlists new_renlists( void ) ;

/***************************************************************************/
/* drop_renlists 							   */
/*									   */
/* Drop h, returning any internal storage and put it on the free list. 	   */
/***************************************************************************/

void drop_renlists( pRenlists h ) ;

/***************************************************************************/
/* store_renlists 							   */
/*									   */
/* Store a renlists pointer in a structure, adjusting the refcnts. 	   */
/***************************************************************************/

void store_renlists( pRenlists *dest, pRenlists src ) ;

/***************************************************************************/
/* new_sexp_sexp 							   */
/*									   */
/* Carve and return a new S-expression sub-expression element and fill it  */
/* in with the parameter.						   */
/***************************************************************************/

pSexp new_sexp_sexp( pSexp s ) ;

/***************************************************************************/
/* new_sexp_bs								   */
/*									   */
/* Carve and return a new S-expression sub-expression element and fill it  */
/* with the given byte string.						   */
/***************************************************************************/

pSexp new_sexp_bs( pBytes b ) ;

/***************************************************************************/
/* new_sexp_kw								   */
/*									   */
/* Carve and return a new keyword S-expression node.			   */
/***************************************************************************/

pSexp new_sexp_kw( int k ) ;

/***************************************************************************/
/* copy_sexp_elt							   */
/*									   */
/* Copy one S-expression element.					   */
/***************************************************************************/

pSexp copy_sexp_elt( pSexp e ) ;

#endif /* __FREE_H__ */
