/***************************************************************************/
/* genlookup.c								   */
/*									   */
/* Program to generate lookup tables for reserved words.		   */
/*									   */
/* Read from stdin a file containing words.				   */
/*									   */
/* The first word is a prefix for defined symbols.			   */
/* Symbols will be forced into upper case.				   */
/* The remaining words are the reserved words.				   */
/*									   */
/* Put out two files: lookup.c and lookup.h				   */
/*									   */
/*  Copyright 1997, Carl Ellison					   */
/*									   */
/*  This file may be copied and used by anyone, provided the		   */
/*  file is kept intact.						   */
/***************************************************************************/

#include <stdio.h>
#include <string.h>
#include <ctype.h>
#include <stdlib.h>

#define NAM_LTH  (80)		/* maximum length name */

typedef char *CP ;

typedef struct _N {
  struct _N *lsr, *gtr ;	/* tree pointers */
  char n[NAM_LTH] ;		/* spelling of the word, as a C string */
} Nam, *pNam ;			/* data type */

#define TNULL  ((pNam) 0)	/* null tree entry */

char pfx[NAM_LTH] ;		/* prefix for defined symbols */

pNam tree = TNULL ;		/* tree of defined names */
int n_names =  0 ;		/* number of names in the tree */

typedef struct {
  char c ;			/* character to be translated */
  char *x ;			/* translation */
} xlate, *pXlate ;

static xlate xtinit[] = {
  { '(', "RPAR" },  { ')', "LPAR" },  { '[', "LSQ" },  { ']', "RSQ" },
  { '{', "LBRACE" },  { '}', "RBRACE" },  { '!', "EXCLAM" },  { '@', "AT" },
  { '*', "STAR" },  { '#', "SPLAT" },  { '$', "DOLLAR" },  { '%', "PCT" },
  { '^', "HAT" },  { '&', "AMP" },  { '_', "UNDER" },  { '=', "EQ" },
  { '+', "PLUS" },  { '\\', "BACKSLASH" },  { '|', "BAR" },  { ';', "SEMI" },
  { ':', "COLON" },  { '\'', "RQUOT" },  { '`', "LQUOT" },  { '~', "TILDE" },
  { '"', "QUOT" },  { ',', "COMMA" },  { '<', "LT" },  { '>', "GT" },
  { '.', "DOT" },  { '/', "SLASH" },  { '?', "QUEST" },
  { '\0', "\n" }		/* end of list is '\0' */
} ;

static CP xt[256] ;		/* translation table */

/***************************************************************************/
/* init_xt								   */
/*									   */
/* Initialize the table, xt[].						   */
/***************************************************************************/

void init_xt(void)
{
  int i ;
  pXlate p = &xtinit[0] ;

  for (i=0; i<256; i++) xt[i] = NULL ;
  while ((p->c) != '\0') {
    xt[ (i=p->c) ] = p->x ;
    p++ ;
  } /* while */

  return ;
} /* init_xt */

/***************************************************************************/
/* compare -- compare two strings, returning a tristate value		   */
/***************************************************************************/

#define EQUAL (0)
#define LESSER (1)
#define GREATER (2)

int compare( char *a, char *b )
{
  char x, y ;

  do {
    x = *(a++) ;
    y = *(b++) ;
    if (x < y)
      return ( LESSER ) ;
    else if (x > y)
      return ( GREATER ) ;
  } while ( x != '\0' ) ;
  return ( EQUAL ) ;
} /* compare */

/***************************************************************************/
/* lookup -- look up a string and plant it in the binary tree.		   */
/***************************************************************************/

void lookup( char *in, pNam *t )
{
  if (in == (char *) 0) return ;
  if ((*t) == TNULL) {
    pNam x = (pNam) malloc( sizeof( Nam ) ) ;
    int l = strlen( in ) ;

    n_names++ ;			/* count the new name */
    if (x == TNULL) return ;
    x->lsr = x->gtr = TNULL ;	/* no sub-tree */
    if (l >= NAM_LTH) l = NAM_LTH - 1 ;
    in[l] = '\0' ;
    memcpy( &(x->n[0]), in, l+1 ) ; /* enter the name spelling */
    (*t) = x ;
  } else
    switch (compare(in, &((*t)->n[0]))) {
    case LESSER: lookup( in, &((*t)->lsr) ) ; return ;
    case GREATER: lookup( in, &((*t)->gtr) ) ; return ;
    case EQUAL: return ;
    } /* switch, else */
} /* lookup */

/***************************************************************************/
/* linearize -- using .gtr as a forward pointer, add to the tail of a list */
/* in global variable l_tail from walking the tree, t.			   */
/***************************************************************************/

pNam l_head = TNULL ;		/* linear list head */
pNam l_tail = TNULL ;		/* linear list tail */

void linearize( pNam t )
{
  if (t != TNULL) {
    pNam g = t->gtr ;

    linearize( t->lsr ) ;
    t->lsr = t->gtr = TNULL ;	/* no other list */

    if (l_tail == TNULL)
      l_head = t ;
    else
      l_tail->gtr = t ;
    l_tail = t ;

    linearize( g ) ;
  } /* if */
} /* linearize */

/***************************************************************************/
/* balance -- produce a balanced tree of N items, from l_head. 		   */
/***************************************************************************/

pNam balance( int N )
{
  int R, L ;			/* right and left counts */
  pNam lt, ct ;			/* left and center trees */

  if (N-- <= 0) return ( TNULL ) ; /* if empty, no tree */
  /* N is now the # for the two sides */
  R = N / 2 ;
  L = N - R ;
  lt = balance( L ) ;		/* get the left tree */
  ct = l_head ;
  l_head = ct->gtr ;		/* take the center element */
  ct->lsr = lt ;		/* put in the lesser tree */
  ct->gtr = balance( R ) ;	/* get the right tree */
  return ( ct ) ;
} /* balance */

/***************************************************************************/
/* do_balance -- balance tree.						   */
/***************************************************************************/

void do_balance( void )
{
  linearize( tree ) ;
  tree = balance( n_names ) ;
  return ;
} /* do_balance */

/***************************************************************************/
/* fputcap -- put s, as all caps					   */
/***************************************************************************/

void fputcap( char *s, FILE *f )
{
  int c ;

  if (s == (char *) 0) return ;
  while ((c = *(s++)) != '\0') {
    c = toupper( c ) ;
    if (xt[c] == NULL) {
      if (c == '-')
	fputc( '_', f ) ;
      else
	fputc( c, f ) ;
    } else {
      fputc( '_', f ) ;
      fputs( xt[c], f ) ;
      fputc( '_', f ) ;
    } /* else -- there's a translation */
  }
  return ;
} /* fputcap */

/***************************************************************************/
/* indent -- put out n spaces.						   */
/***************************************************************************/

void indent( int n, FILE *f )
{
  while ((n--) > 0) fputc( ' ', f ) ;
  return ;
} /* indent */

/***************************************************************************/
/* write_h -- write lookup.h into f					   */
/***************************************************************************/

int value = 1 ;			/* value of the next symbol */

void write_h( pNam t, FILE *f )
{
  if (t == TNULL) return ;
  write_h( t->lsr, f ) ;
  fputs( "#define ", f ) ;
  fputcap( pfx, f ) ;
  fputcap( t->n, f ) ;
  fprintf( f, " (%d)\n", value++ ) ;
  write_h( t->gtr, f ) ;
  return ;
} /* write_h */

/***************************************************************************/
/* write_s -- write lookup.h into f as a list of word spellings.	   */
/***************************************************************************/

void write_s( pNam t, FILE *f )
{
  if (t == TNULL) return ;
  write_s( t->lsr, f ) ;
  fprintf( f, ",\n\
  { %d, \"%s\" }", strlen( t->n ), t->n ) ;
  write_s( t->gtr, f ) ;
  return ;
} /* write_s */

/***************************************************************************/
/* write_c -- write lookup.c into f					   */
/***************************************************************************/

void write_c( int n, pNam t, FILE *f )
{
  if (t == TNULL) {
    indent( n, f ) ;
    fputs( "return (-1) ;   /* undefined */\n", f ) ;
    return ;
  } /* no tree here */

  indent( n, f ) ;
  fprintf( f, "switch ( compare( x, xl, \"%s\" ) ) {\n", t->n ) ;

  indent( n, f ) ;
  fputs( "case LESSER:\n", f ) ;
  write_c( n+2, t->lsr, f ) ;

  indent( n, f ) ;
  fputs( "case EQUAL:  return ( ", f ) ;
  fputcap( pfx, f ) ;
  fputcap( t->n, f ) ;
  fputs( " ) ;\n", f ) ;

  indent( n, f ) ;
  fputs( "case GREATER:\n", f ) ;
  write_c( n+2, t->gtr, f ) ;

  indent( n, f ) ;
  fprintf( f, "} /* compare( x, xl, \"%s\" ) */\n", t->n ) ;
  return ;
} /* write_c */

/***************************************************************************/
/* main program								   */
/***************************************************************************/

int main( void )
{
  char symbol[NAM_LTH] ;
  FILE *f ;			/* output file */

  init_xt() ;			/* initialize the translation table */


  scanf( "%s", pfx ) ;		/* read the prefix */

  while ( scanf( "%s", symbol ) != EOF )
    lookup( symbol, &tree ) ;

  do_balance() ;

  f = fopen( "lookup.c", "w" ) ;
  fputs( "#include \"lookup.h\"\n\
#include \"compare.h\"\n\
\n\
int lookup( char *x, int xl )\n\
{\n\
", f ) ;
  write_c(4,tree,f) ;		/* write the .c file */
  fputs( "  return (-1) ;  /* needed by some compilers */\n\
} /* lookup */\n\
", f ) ;
  fclose(f) ;

  f = fopen( "lookup.h", "w" ) ;
  fputs( "/* lookup.h -- generated automatically */\n\
\n\
#if !defined(__LOOKUP_H__)\n\
#define __LOOKUP_H__\n\
\n\
int lookup( char *x, int xl ) ;	/* look up a reserved word */\n\
\n", f ) ;
  write_h(tree,f) ;		/* write the .h file */
  fputs( "\n\
#endif /* __LOOKUP_H__ */\n", f ) ;
  fclose(f) ;

  /** main  ********  continued **/

  f = fopen( "wordtab.c", "w" ) ;
  fprintf( f, "/* wordtab.c -- generated automatically */\n\
\n\
#include \"wordtab.h\"\n\
\n\
Spelling spell[%d] = {\n\
  { 3, \"???\" }", (n_names+1) ) ;

  write_s( tree, f ) ;

  fputs( "\n} ;\n", f ) ;

  fclose( f ) ;

  return (0) ;
} /* main */
