/***************************************************************************/
/* psexp.c								   */
/*									   */
/*  Routines for parsing S-expressions from an in-memory block. 	   */
/***************************************************************************/

#include <values.h>
#include "strread.h"
#include "abbrev.h"
#include "sexp.h"
#include "free.h"
#include "errcodes.h"
#include "lookup.h"

/***************************************************************************/
/* p_lth								   */
/*									   */
/* Read the length field from the current string reader.		   */
/***************************************************************************/

int p_lth( ULP pRes, pSreader s )
{
  UL res = 0 ;			/* result */
  UC ch ;			/* current character */

  if ((pRes == NULL) || (s == NULL))
    return ( SPKI_ERR_NULL_PARAM ) ;

  ch = STRR_READ( s ) ;		/* get the first character */
  if ((ch < '1') || (ch > '9')) return ( SPKI_ERR_BAD_LTH_FORMAT ) ;

  /* incorporate the current ch and then get a new one */
  while (STRR_MORE( s ) && ('0' <= ch) && (ch <= '9')) {
    if (res > (MAXLONG/10)) return ( SPKI_ERR_LTH_TOO_BIG ) ;
    res = res*10 + (ch - '0') ;
    ch = STRR_READ( s ) ;	/* get the next character */
  } /* while */

  if (ch != ':') return ( SPKI_ERR_COLON_EXPECTED ) ;
  *pRes = res ;			/* return the value */
  return ( 0 ) ;		/* and no error */
} /* p_lth */

/***************************************************************************/
/* p_bytes								   */
/*									   */
/* Parse a byte string from the current string reader and pack it into	   */
/* a new byte string block.						   */
/***************************************************************************/

int p_bytes( pBytes *ppb, pSreader s )
{
  /* stack variable -- must be dropped on return */
  pBytes pb = NULL ;		/* block we're building */

  UC ch ;			/* working character */
  UL strlen ;			/* string length */
  int r = 0 ;			/* return value */

  if ((ppb == NULL) || (s == NULL))
    return ( SPKI_ERR_NULL_PARAM ) ;
  store_bytes( ppb, NULL ) ;	/* no block for output yet */

  pb = new_bytes() ;		/* comes with refcnt == 1, for the stack */
  if (pb == NULL) return ( SPKI_ERR_NO_MEM ) ;
  pb->drop_cont = 0 ;		/* we're not carving memory for strings */

  ch = STRR_LOOK( s ) ;		/* look at the first character */
  if (ch == '[') {		/* read a display hint */
    ch = STRR_READ( s ) ;	/* read and drop it */
    r = p_lth( &strlen, s ) ;	/* read the length field */
    if (r == 0) {
      pb->d_lth = strlen ;	/* note the display hint length */
      pb->disp = s->cur ;		/* point to the string in memory */
      s->cur += strlen ;		/* bump the current pointer */
      ch = STRR_READ( s ) ;	/* read the closing character */
      if (ch != ']') r = SPKI_ERR_DISP_HINT_NOT_CLOSED ;
      if (!STRR_MORE(s)) r = SPKI_ERR_DISP_HINT_TOO_LONG ;
    }
  } /* if display hint */

  if (r == 0) {
    r = p_lth( &strlen, s ) ;	/* get the string length */
    if (r == 0) {
      pb->v_lth = strlen ;	/* plant the string length */
      pb->val = s->cur ;		/* and the string pointer */
      s->cur += strlen ;		/* and advance that pointer */
      if (s->cur > s->end) r = SPKI_ERR_STRING_TOO_LONG ;
    }
  } /* read string itself */

  if (r == 0) {
    if (pb->d_lth != 0)
      pb->lu = -1 ;		/* no lookup value unless naked string */
    else pb->lu = lookup( pb->val, pb->v_lth ) ;
    store_bytes( ppb, pb ) ;	/* plant the block in storage */
  } /* if */

  drop_bytes( pb ) ;		/* drop the stack reference */
				/* and maybe the block */

  return ( r ) ;		/* and return any error code */
} /* p_bytes */

/***************************************************************************/
/* p_sexp								   */
/*									   */
/* Parse an S-expression from a string block.				   */
/* An S-expression is enclosed in "()" and starts with a byte string.	   */
/* Look up the byte string it starts with.				   */
/***************************************************************************/

int p_sexp( pSexp *pps, pSreader s )
{
  /* stack storage of pointers -- must be dropped on return */
  pSexp ps = NULL;		/* new block for this S-expression */
  pSexp ps_tail = NULL ;	/* tail of the list */
  pSexp pt = NULL ;		/* new term */
  pSexp psub = NULL ;		/* sub-expression */
  pBytes pb = NULL ;		/* byte string to start the expression */

  UC ch ;			/* working character */
  int r ;			/* return value */
  UCP expr_start ;		/* start of the expression */

  if ((pps == NULL) || (s == NULL))
    return ( SPKI_ERR_NULL_PARAM ) ;
  store_sexp( pps, NULL ) ;	/* no return value yet */

  expr_start = s->cur ;		/* note the start of the expression */
  ch = STRR_READ( s ) ;		/* get the 1st char -- better be '(' */
  if (ch != '(') return ( SPKI_ERR_BAD_SEXP_START ) ;

  r = p_bytes( &pb, s ) ;	/* get the initial byte string */
  if (r != 0) {			/* unless there's an error */
    drop_bytes( pb ) ;		/* drop any block returned */
    return ( r ) ;		/* and return the error code */
  } /* if */

  /* OK, have the initial byte string -- look it up and start */
  /*  expression block. */

  ps = new_sexp() ;		/* comes with refcnt==1, for the stack */
  if (ps == NULL) return ( SPKI_ERR_NO_MEM ) ;
  ps->term = NULL ;
  ps->expb = expr_start ;	/* note start of the expression */

  if (pb->lu > 0) {		/* the keyword lookup value */
    ps->typ = S_KW ;		/* known keyword -- known object type */
    ps->u.k = pb->lu ;		/* plant that keyword */
    /* no string block stored, in this case */
  } else {
    ps->typ = S_UKW ;		/* unknown keyword, start of S-expression */
    store_bytes( &(ps->u.b), pb ) ; /* plant the byte string block */
  }

  store_sexp( &ps_tail, ps ) ;	/* note current tail of the list */

  /* We have an S-expression start, but now we need to get the rest of it. */

  /* p_sexp  continued */

  r = 0 ;			/* flag for breaking the loop */
  while ((ps != NULL) && (r == 0) && STRR_MORE( s )) {
    ch = STRR_LOOK( s ) ;	/* get the next character */
    switch (ch) {
    case ')':
      ch = STRR_READ( s ) ;	/* read it */
      ps->explth = s->cur - expr_start ; /* calculate the total length */
      store_sexp( pps, ps ) ;	/* store this one away */
      store_sexp( &ps, NULL ) ;	/* signal that we're successful */
      break ;			/* break the loop to return */

    case '1': case '2': case '3': case '4': case '5': 
    case '6': case '7': case '8': case '9': case '[':
      r = p_bytes( &pb, s ) ;	/* read the byte string */
      if (r != 0) break ;	/* if error, stop this loop */
      /* OK, carve a new term block for this byte string */
      drop_sexp( pt ) ;
      pt = new_sexp() ;		/* new term block  */
      if (pt == NULL)
	r = SEXP_ERR_NO_MEM ;
      else {
	store_sexp( &(ps_tail->term), pt ) ;
	store_sexp( &ps_tail, pt ) ;
	pt->typ = S_BS ;	/* byte string type */
	store_bytes( &(pt->u.b), pb ) ;	/* the string */
      } /* else */
      break ;

    case '(':			/* sub-expression */
      r = p_sexp( &psub, s ) ;	/* read the sub-expression */
      if (r != 0) break ;	/* if error, break out */
      /* OK, get a new term block */
      drop_sexp( pt ) ;
      pt = new_sexp() ;		/* new term block  */
      if (pt == NULL)
	r = SEXP_ERR_NO_MEM ;
      else {
	store_sexp( &(ps_tail->term), pt ) ;
	store_sexp( &ps_tail, pt ) ;
	pt->typ = S_SEXP ;	/* sub-expression type */
	store_sexp( &(pt->u.s), psub ) ; /* the sub-expression */
      } /* else */
      break ;

    default:  r = SPKI_ERR_BAD_SEXP_FORMAT ;
    } /* switch */
  } /* while more string and no error */

  if (ps != NULL)		/* some error condition? */
    if (r == 0)			/* if none indicated */
      r = SPKI_ERR_SEXP_NOT_CLOSED ; /* then we're out of string */

  /* drop all the stack references */
  drop_bytes( pb ) ;		/* if any... */
  drop_sexp( ps ) ;		/* mostly, clean up refcnt */
  drop_sexp( ps_tail ) ;	/* and maybe drop a block */
  drop_sexp( pt ) ;		/* or two */
  drop_sexp( psub ) ;		/* ... */

  return ( r ) ;		/* and return */
} /* p_sexp */

/***************************************************************************/
/* kill_expb								   */
/*									   */
/* Walk s recursively, killing all expb and explth fields.		   */
/* This expression started from a contiguous block (e.g. a literal string) */
/*     but it is being edited into something non-contiguous.		   */
/***************************************************************************/

void kill_expb( pSexp s )
{
  while (s != NULL) {
    s->expb = NULL ;
    s->explth = 0 ;
    if (s->typ == S_SEXP) kill_expb( s->u.s ) ;	/* walk sub-expressions */
    s = s->term ;
  } /* while */
  return ;
} /* kill_expb */

/***************************************************************************/
/* lit_sexp								   */
/*									   */
/* Parse and return a literal S-expression.				   */
/***************************************************************************/

int lit_sexp( pSexp *pps, char *lit )
{
  Sreader rdr ;
  pSexp out = NULL ;		/* the result expression */
  int r ;			/* return value */

  if ((pps == NULL) || (lit == NULL)) return ( SPKI_ERR_NULL_PARAM ) ;
  *pps = NULL ;
  rdr.cur = rdr.top = lit ;	/* point to the literal string */
  rdr.end = rdr.cur + strlen( lit ) ; /* and the end of the string */
  r = p_sexp( &out, &rdr ) ;
  if (r != 0) return ( r ) ;
  kill_expb( out ) ;		/* drop all expb indications */
  *pps = out ;			/* return the new expression */
  return ( 0 ) ;		/* and give no error */
} /* lit_sexp */
