/***************************************************************************/
/* sutil.c								   */
/*									   */
/* Small S-expression utilities.					   */
/***************************************************************************/

#include "sutil.h"
#include "abbrev.h"
#include "sexp.h"
#include "lookup.h"
#include "errcodes.h"
#include "tuple.h"
#include "bs_compare.h"

/***************************************************************************/
/* pull_typ								   */
/*									   */
/* Pull the .typ field from an Sexp and return it.			   */
/***************************************************************************/

int pull_typ( pSexp s )
{
  if (s == NULL)
    return ( SPKI_ERR_NULL_PARAM ) ;
  else
    return ( s->typ ) ;
} /* pull_typ */

/***************************************************************************/
/* pull_resword								   */
/*									   */
/* Pull any reserved word from the S-expression node at s.		   */
/* If s is a sub-expression, pull its reserved word.			   */
/***************************************************************************/

int pull_resword( pSexp s )
{
  switch (pull_typ( s )) {
  case S_KW: return ( s->u.k ) ;
  case S_UKW: return ( -1 ) ;	/* unknown */
  case S_BS: return (s->u.b->lu ) ;
  case S_SEXP: return ( pull_resword( s->u.s ) ) ;
  default:
  } /* switch */
  return ( -1 ) ;		/* unknown */
} /* pull_resword */

/***************************************************************************/
/* pull_hashval								   */
/*									   */
/* Pull the fields from a (hash...) S-expression.			   */
/***************************************************************************/

int pull_hashval( pSexp h, UCP *hvp, UI *hvl, int *alg )
{
  pBytes b ;

  *hvl = 0 ;			/* prepare for a null return */
  *hvp = NULL ;
  *alg = -1 ;			/* unknown alg */

  if (pull_resword( h ) != O_HASH) return ( SPKI_ERR_NOT_HASH_SEXP ) ;
  h = h->term ;			/* go to the next term */
  *alg = pull_resword( h ) ;	/* get the hash algorithm */
  if (*alg <= 0) return ( SPKI_ERR_UNKNOWN_HASH ) ;
  h = h->term ;			/* next term is hash value */
  if (h == NULL) return ( SPKI_ERR_BAD_HASH_SEXP ) ;
  if (h->typ != S_BS) return ( SPKI_ERR_BAD_HASH_SEXP ) ;
  b = h->u.b ;			/* the byte string */
  *hvp = b->val ;		/* return the hash value */
  *hvl = b->v_lth ;		/* and its length */
  return ( 0 ) ;		/* and give a normal return */
} /* pull_hashval */

/***************************************************************************/
/* pull_subject_type							   */
/*									   */
/* Pull a subject type out of the subject in a tuple.			   */
/***************************************************************************/

int pull_subject_type( pTuple t )
{
  pSexp s = NULL ;		/* the subject (walker, no drop) */
  int st = -1 ;			/* subject type */

  if (t == NULL) return ( SPKI_ERR_NULL_PARAM ) ;
  s = t->s ;			/* get the subject */
  switch (st = pull_resword( s )) {
  case O_SUBJECT:
    s = s->term ;		/* go to the next term */
    if (pull_typ( s ) != S_SEXP) return (-1) ; /* must be sub-expression */
    return ( pull_resword( s->u.s ) ) ; /* it's the type of the 1st term */
  default:
  } /* switch */
  return ( st ) ;
} /* pull_subject_type */

/***************************************************************************/
/* cmp_hash_1								   */
/*									   */
/* Compare a hash S-expression to hash values in the argument stream.	   */
/***************************************************************************/

int cmp_hash_1( pSexp h, UCP hvp, UI hvl, int alg )
{
  UCP hvp1 ;
  UI hvl1 ;
  int alg1 ;
  int r ;

  r = pull_hashval( h, &hvp1, &hvl1, &alg1 ) ;
  if (r != 0) return ( r ) ;
  if (alg1 != alg) return ( SPKI_ERR_DIFF_HASH_ALGS ) ;
  if (hvl1 != hvl) return ( SPKI_ERR_DIFF_HASH_LTHS ) ;

  return ( -1 ) ;		/* this function hasn't been finished yet */
} /* cmp_hash_1 */

/***************************************************************************/
/* equal_sexp								   */
/*									   */
/* Compare two S-expressions and return 1 if they're equal.		   */
/***************************************************************************/

int equal_sexp( pSexp a, pSexp b )
{
  if (a == b) return ( 1 ) ;
  if ((a == NULL) || (b == NULL)) return ( 0 ) ;
  while ( (a != NULL) && (b != NULL) ) {
    int t = pull_typ( a ) ;
    if (t != pull_typ( b )) return ( 0 ) ;
    switch (t) {
    case S_KW:
      if ( a->u.k != b->u.k ) return ( 0 ) ;
      break ;
    case S_UKW: case S_BS:
      if ( bs_compare( a->u.b, b->u.b ) != EQUAL ) return ( 0 ) ;
      break ;
    case S_SEXP:
      if (! equal_sexp( a->u.s, b->u.s ) ) return ( 0 ) ;
      break ;
    default:
      return ( 0 ) ;		/* don't know this kind of block */
    } /* switch */

    a = a->term ;
    b = b->term ;
  } /* while */
  return ( b == a ) ;		/* equal only if a and b end together */
} /* equal_sexp */

/***************************************************************************/
/* equal_tuple								   */
/*									   */
/* Compare two tuples and see if they're equal.  If so, return 1.  O'E, 0. */
/***************************************************************************/

int equal_tuple( pTuple a, pTuple b )
{
  if (a == b) return ( 1 ) ;
  if ((a == NULL) || (b == NULL)) return ( 0 ) ;
  if (a->active != b->active) return ( 0 ) ;
  if (a->d != b->d) return ( 0 ) ;
  if (a->j != b->j) return ( 0 ) ;
  if (a->ik != b->ik) return ( 0 ) ;
  if (a->inam != b->inam)
    if (bs_compare( a->inam, b->inam ) != EQUAL) return ( 0 ) ;
  if (a->sk != b->sk) return ( 0 ) ;
  if (a->snam != b->snam)
    if (bs_compare( a->snam, b->snam ) != EQUAL) return ( 0 ) ;
  if (!equal_sexp( a->rnames, b->rnames )) return ( 0 ) ;
  if (a->th != b->th) return ( 0 ) ;
  if (!equal_sexp( a->s, b->s )) return ( 0 ) ;
  if (!equal_sexp( a->t, b->t )) return ( 0 ) ;
  if (bs_compare( a->nb, b->nb ) != EQUAL) return ( 0 ) ;
  if (bs_compare( a->na, b->na ) != EQUAL) return ( 0 ) ;
  return ( 1 ) ;		/* tuples are equal */
} /* equal_tuple */
