/***************************************************************************/
/* tuple.h								   */
/*									   */
/* Header file defining data structures and calls for tuple.c.		   */
/* These support the tuple and name reduction for SPKI.			   */
/*									   */
/*  See free.h for text on the handling of refcnt.			   */
/*									   */
/***************************************************************************/

#if !defined(__TUPLE_H__)
#define __TUPLE_H__

#include "hashtabs.h"		/* define MAXHASHLTH, etc. */
#include "sexp.h"		/* s-expression structure definitions */
#include "abbrev.h"		/* type abbreviations */
#include "sha.h"		/* define SHA structures */
#include "md5.h"		/* define MD5 structures */

/***************************************************************************/
/* Forward references to all structures defined here. 			   */
/***************************************************************************/

typedef struct _thresh_ *pThresh ; /* Threshold rendezvous block  */
typedef struct _tuple_ *pTuple ; /* tuple (or piece of threshold) */
typedef struct _pkh_ *pPkh ;	/* public-key hash rendezvous */
typedef struct _ren5_ *pRen5 ;	/* rendezvous hash node */
typedef struct _tred_ *pTred ;	/* Threshold subject reduction node */
typedef struct _renlists_ *pRenlists ; /* 5-tuple rendezvous lists */

/***************************************************************************/
/* Define a threshold rendezvous structure.				   */
/*									   */
/*  This is for threshold subjects -- k-of-n subjects, such that k of them */
/*  must eventually reduce to the same subject in order to propagate.	   */
/*  Note, however, that it is possible for a threshold subject to	   */
/*  propagate to multiple descendents.  Therefore, this rendezvous node	   */
/*  might yield multiple children. 					   */
/***************************************************************************/

typedef struct _thresh_ {
  pThresh nxt ;			/* for lists */
  int refcnt ;			/* reference count */
  short k ;			/* # of elts that have to match to succeed */
  short j ;			/* index into the threshold */
				/* j<0: ACL; j==0: ik; j>0: th */
  pThresh th ;			/* threshold block */
  pRenlists ik ;		/* issuer key (or hash of it) */
  int n ;			/* # of slots in the array (max # subjects) */
  pTred *tr ;			/* array of tuple lists waiting to reduce */
} Thresh ;

/***************************************************************************/
/* Define a tuple structure						   */
/***************************************************************************/

typedef struct _tuple_ {
  pTuple nxt ;			/* for the free list */
  int refcnt ;			/* reference count.  On 0, reclaim it. */
  UC active ;			/* boolean: this tuple is active */
  UC d ;			/* boolean: permission to delegate */
  short j ;			/* index into the threshold or issuer */
				/* j<0: ACL; j==0: ik; j>0: th */
  /* start with internal references, pulled out for easy access */
  pRenlists ik ;		/* issuer key (or hash of it) */
  pBytes inam ;			/* issuer name (if any) */
  pRenlists sk ;		/* subject key (or hash of it) */
  pBytes snam ;			/* 1st subject name (if any) */
  pSexp rnames ;		/* remaining names */
  pThresh th ;			/* threshold block */
  pSexp s ;			/* subject */
  pSexp t ;			/* tag */
  pBytes nb ;			/* not_before date.time */
  pBytes na ;			/* not_after  date.time */
} Tuple ;

/***************************************************************************/
/*									   */
/*   Hash Tables							   */
/*									   */
/*   Some things are hashed, for rendezvous purposes. 			   */
/*									   */
/*   1.  public keys: hashed by all algorithms implemented,		   */
/*             currently MD5 and SHA-1					   */
/*    These are matched against (hash ...) blocks.			   */
/*    The hashing process also creates a single Renlist node for each	   */
/*    key (or hash that has no key) where rendezvous lists are held.	   */
/*    For ACLs (with no issuer), there is a single Renlists node	   */
/*    kept in tuple.c							   */
/*									   */
/*   2.  (key, name) for name reduction purposes			   */
/*									   */
/*   There is one hash table for each of these things.  Each hash	   */
/*   table has a native data block.  These are defined below.		   */
/***************************************************************************/

/***************************************************************************/
/* Renlists								   */
/*									   */
/* Lists of ren5 blocks for this object -- either a key or a hash.	   */
/***************************************************************************/

typedef struct _renlists_ {
  pRenlists nxt ;		/* for free storage */
  int refcnt ;			/* reference count, drop on 0 */
  HashVal hash ;		/* preferred hash for this object */
  pSexp obj ;			/* the hashed object (for result report) */
  pRen5 i ;			/* issuers for 5-tuple reduction */
  pRen5 s ;			/* subjects for 5-tuple reduction */
} Renlists ;

/***************************************************************************/
/* Pkh									   */
/*									   */
/* This is used for the public key hash tables.				   */
/* It is kept around to stand for the hashed object (key or hash).	   */
/***************************************************************************/

typedef struct _pkh_ {
  pPkh nxt ;			/* for chaining and free storage */
  HashVal hash ;		/* hash of this key or from a hash block */
  pRenlists obj ;		/* rendezvous lists for this object */
} Pkh ;

/***************************************************************************/
/* Ren5									   */
/*									   */
/* This is used in Renlists, for 5-tuple reduction.			   */
/***************************************************************************/

typedef struct _ren5_ {
  pRen5 nxt ;			/* for chaining and free storage */
  HashVal hash ;		/* its hash */
  pTuple t ;			/* the tuple using that object, if any */
} Ren5 ;

/***************************************************************************/
/* Tred									   */
/*									   */
/* This is used for threshold reduction and hangs off a threshold block.   */
/* The only tuples that get Tred blocks are those that are active and	   */
/* do not have subject names.						   */
/***************************************************************************/

typedef struct _tred_ {
  pTred nxt ;			/* for chaining and free lists */
  pThresh th ;			/* the threshold block of which we're part */
  pRenlists s ;			/* t's subject */
  pTuple t ;			/* my tuple */
} Tred ;

/***************************************************************************/
/* Documentation of the rendezvous and reduction process.		   */
/***************************************************************************/

/***************************************************************************/
/* 1.  Public key hashing						   */
/*									   */
/*  Each public key gets entered in this table.  There is no need for 	   */
/*  a tuple entry for that public key.  If a key already exists in the 	   */
/*  table, then we will point to that key's Renlists rather than make a    */
/*  new one, to make sure that all identical keys reduce properly.	   */
/*									   */
/***************************************************************************/

/***************************************************************************/
/* 2. Hash lookup							   */
/*									   */
/*  After all keys have been entered, this table is scanned for any hash   */
/*  block encountered.  If the hash value is in the table (implying the key*/
/*  has been seen) then that hash gets the Renlists for that key.	   */
/*  Otherwise, a new Renlists block is created for that hash.		   */
/*  Hashes with no public key should not be able to rendezvous, but	   */
/*  we won't deal with that here.					   */
/***************************************************************************/

/***************************************************************************/
/* 3.  Name reduction hash table entries.				   */
/*									   */
/*  One of these blocks is used for rendezvous of name definitions	   */
/*  and names as subjects.						   */
/*									   */
/*  A name subject can be of the form (name K_0 N_0 N_1 N_2 N_3 ...).	   */
/*  The name reduction will replace (K_0 N_0) with (K_1)		   */
/*  so that the name becomes (name K_1 N_1 N_2 N_3 ...).		   */
/*									   */
/*  Therefore, we hash only on (K_0 N_0) for this entry, even if	   */
/*  the name block is much longer.					   */
/*									   */
/*  The hash is over the public key and the name spelling.  The		   */
/*  public key is an S-expression.  The name is a byte string.		   */
/*  If the name has a display hint, that is hashed also.		   */
/*									   */
/*  As with the previous definition, there are two parallel hash tables	   */
/*  and they are both searched for each new entry.			   */
/*									   */
/*  If the new entry is already listed in its own hash table, then	   */
/*  it doesn't need to be added again.					   */
/*									   */
/*  If it does need to be added, then it is checked against the opposite   */
/*  gender, to see if a reduction is possible.				   */
/*									   */
/***************************************************************************/

/***************************************************************************/
/* 4.  Threshold subject reduction.					   */
/*									   */
/*   A threshold subject is a small tree of normal tuples.  At first,	   */
/*   all tuples in that tree are the same except for the subject.	   */
/*   They all point back to a threshold "join" structure: Thresh.	   */
/*   However, after name and 5-tuple reduction proceed, some of these	   */
/*   subjects might be the same while some of their authorizations and	   */
/*   validity periods might differ.					   */
/*									   */
/*   The threshold subject reduction rendezvous node exists to help	   */
/*   locate when multiple different branches of the threshold tree	   */
/*   have the same subject at the end.  As soon as we discover more	   */
/*   than K of these different branches with the same subject, we	   */
/*   create a non-threshold subject (matching that subject) with	   */
/*   authorization and validity that are the intersection of those	   */
/*   in the K that we discovered.					   */
/*									   */
/*   If there are more than K which match, we must take all subsets	   */
/*   of size K and enter the resulting tuples into the process.		   */
/*									   */
/***************************************************************************/

/***************************************************************************/
/* 5-tuple reduction.							   */
/*									   */
/* Normal 5-tuples have issuers and subjects that are single keys.	   */
/* (A left-most 5-tuple will be from an ACL and have a NULL issuer.	   */
/* A right-most 5-tuple might have a subject that is a hash that was	   */
/* never matched to a public key.)					   */
/*									   */
/* Whenever a new 5-tuple is created, it is entered in the appropriate	   */
/* lists.  There are usually two Renlists nodes, one for the issuer	   */
/* and one for the subject.  The 5-tuple is entered in the issuer list	   */
/* of the issuer and the subject list of the subject, but if it is active  */
/* it is checked against the other list in each case, looking for a match  */
/* that could yield a reduction.					   */
/*									   */
/* If a reduction is made, that new 5-tuple is entered (recursively)	   */
/* in its lists.							   */
/***************************************************************************/

/***************************************************************************/
/* Calling sequence of the one available function.			   */
/***************************************************************************/

/***************************************************************************/
/* test_hash_walk							   */
/*									   */
/* Test walk_hash against mpl_hash					   */
/*									   */
/*  Hash the same Sexp 3 ways:						   */
/*  1. full contiguous							   */
/*  2. top level expb killed						   */
/*  3. all expb's killed						   */
/***************************************************************************/

void test_hash_walk( void ) ;

/***************************************************************************/
/* do_reduction								   */
/*									   */
/* Reduce a set of credentials.						   */
/*									   */
/*  Note: this code will edit the parameters, replacing some S-expressions */
/*  with equivalent expressions.  (Some hashes become public-keys and	   */
/*  some hashes or keys stay the same, but defined in a common block,	   */
/*  so that content equality can be tested by testing pointer equality.)   */
/*									   */
/*									   */
/* Input is:								   */
/* acl: the current ACL							   */
/* cred: certificate bodies and public keys				   */
/* lim: test limitations						   */
/*    subject: a single subject, presumably the one asking for service	   */
/*    tag: the thing that subject is asking to do			   */
/*    now: the current date and time					   */
/* Any or all of the limitations can be missing, in case we want general   */
/*    results rather than a specific answer to one request. 		   */
/*									   */
/* The result of this process is a list of S-expressions, possibly	   */
/* empty.  These will look like an ACL.  Tuples make it into the result	   */
/* list only if they  are ACL-type (have no issuer), are active and have   */
/* the desired subj (if any) as their subject. 				   */
/*									   */
/***************************************************************************/

int do_reduction( pSexp acl,	/* the current ACL */
		  pSexp cred,	/* the credentials being provided */
		  pSexp subj,	/* limitation: a subject */
		  pSexp tag,	/* limitation: a tag */
		  pBytes now,	/* limitation: the current date&time */
		  pSexp *aout	/* output results */
		) ;

#endif /* __TUPLE_H__ */
